from typing import Any, Sequence

import numpy as np
import scipy.sparse as sp

from . import cpphelpers as lib
from .types import NumberTypes

__author__ = "ltla, jkanche"
__copyright__ = "ltla, jkanche"
__license__ = "MIT"


class TatamiNumericPointer:
    """Initialize a Tatami Numeric Ponter object."""

    def __init__(self, ptr: "Mattress", obj: Any):
        """Initialize the class.

        Args:
            ptr (Mattress): Pointer to a Mattress instance wrapping a tatami Matrix.
            obj (Any): Arbitrary Python object that is referenced by the tatami instance.
                This is stored here to avoid garbage collection.
        """
        self.ptr = ptr
        self.obj = obj

    def __del__(self):
        """Free the reference."""
        lib.free_mat(self.ptr)

    def nrow(self) -> int:
        """Get number of rows.

        Returns:
            int: Number of rows.
        """
        return lib.extract_nrow(self.ptr)

    def ncol(self) -> int:
        """Get number of columns.

        Returns:
            int: Number of columns.
        """
        return lib.extract_ncol(self.ptr)

    def sparse(self) -> bool:
        """Is the matrix sparse?

        Returns:
            bool: True if matrix is sparse.
        """
        return lib.extract_sparse(self.ptr) > 0

    def row(self, r: int) -> Sequence[NumberTypes]:
        """Access a row from the tatami matrix.

        Args:
            r (int): Row to access.

        Returns:
            Sequence[NumberTypes]: Row from the matrix.
        """
        output = np.ndarray((self.ncol(),), dtype="float64")
        lib.extract_row(self.ptr, r, output.ctypes.data)
        return output

    def column(self, c: int) -> Sequence[NumberTypes]:
        """Access a column from the tatami matrix.

        Args:
            c (int): Column to access.

        Returns:
            Sequence[NumberTypes]: Column from the matrix.
        """
        output = np.ndarray((self.nrow(),), dtype="float64")
        lib.extract_column(self.ptr, c, output.ctypes.data)
        return output

    @classmethod
    def from_dense_matrix(cls, x: np.ndarray) -> "TatamiNumericPointer":
        """Initialize class from a dense matrix.

        Args:
            x (np.ndarray): Input numpy array with 2 dimensions.

        Returns:
            TatamiNumericPointer: Instance of the class.
        """

        if len(x.shape) != 2:
            raise ValueError("'x' should be a 2-dimensional array")

        byrow = None
        if x.flags["C_CONTIGUOUS"]:
            byrow = True
        elif x.flags["F_CONTIGUOUS"]:
            byrow = False
        else:
            # I don't think it's possible to hit this, as a (non-view) ndarray
            # should be contiguous in at least one direction.
            raise ValueError("'x' must have contiguous storage for its arrays")

        return cls(
            ptr=lib.initialize_dense_matrix(
                x.shape[0],
                x.shape[1],
                str(x.dtype).encode("utf-8"),
                x.ctypes.data,
                byrow,
            ),
            obj=x,
        )

    @classmethod
    def from_csc_array(cls, x: sp.csc_array) -> "TatamiNumericPointer":
        """Initialize class from a compressed sparse column matrix.

        Args:
            x (scipy.sparse.csc_array): Input sparse matrix.

        Returns:
            TatamiNumericPointer: Instance of the class.
        """

        tmp = x.indptr.astype(np.uint64)

        return cls(
            ptr=lib.initialize_compressed_sparse_matrix(
                x.shape[0],
                x.shape[1],
                len(x.data),
                str(x.data.dtype).encode("UTF-8"),
                x.data.ctypes.data,
                str(x.indices.dtype).encode("UTF-8"),
                x.indices.ctypes.data,
                tmp.ctypes.data,
                False,
            ),
            obj=[tmp, x],
        )

    @classmethod
    def from_csr_array(cls, x: sp.csr_array) -> "TatamiNumericPointer":
        """Initialize class from a compressed sparse row matrix.

        Args:
            x (scipy.sparse.csc_array): Input sparse matrix.

        Returns:
            TatamiNumericPointer: Instance of the class.
        """

        tmp = x.indptr.astype(np.uint64)
        return cls(
            ptr=lib.initialize_compressed_sparse_matrix(
                x.shape[0],
                x.shape[1],
                len(x.data),
                str(x.data.dtype).encode("UTF-8"),
                x.data.ctypes.data,
                str(x.indices.dtype).encode("UTF-8"),
                x.indices.ctypes.data,
                tmp.ctypes.data,
                True,
            ),
            obj=[tmp, x],
        )



# DO NOT MODIFY: this is automatically generated by the ctypes-wrapper

import os
import ctypes as ct

def catch_errors(f):
    def wrapper(*args):
        errcode = ct.c_int(0)
        errmsg = ct.c_char_p(0)
        output = f(*args, ct.byref(errcode), ct.byref(errmsg))
        if errcode.value != 0:
            msg = errmsg.value.decode('ascii')
            lib.free_error_message(errmsg)
            raise RuntimeError(msg)
        return output
    return wrapper

# TODO: surely there's a better way than whatever this is.
dirname = os.path.dirname(os.path.abspath(__file__))
contents = os.listdir(dirname)
lib = None
for x in contents:
    if x.startswith('core') and not x.endswith("py"):
        lib = ct.CDLL(os.path.join(dirname, x))
        break

if lib is None:
    raise ImportError("failed to find the core.* module")

lib.free_error_message.argtypes = [ ct.POINTER(ct.c_char_p) ]

lib.py_extract_column.argtypes = [
    ct.c_void_p,
    ct.c_int,
    ct.c_void_p,
    ct.POINTER(ct.c_int),
    ct.POINTER(ct.c_char_p)
]

lib.py_extract_ncol.restype = ct.c_int
lib.py_extract_ncol.argtypes = [
    ct.c_void_p,
    ct.POINTER(ct.c_int),
    ct.POINTER(ct.c_char_p)
]

lib.py_extract_nrow.restype = ct.c_int
lib.py_extract_nrow.argtypes = [
    ct.c_void_p,
    ct.POINTER(ct.c_int),
    ct.POINTER(ct.c_char_p)
]

lib.py_extract_row.argtypes = [
    ct.c_void_p,
    ct.c_int,
    ct.c_void_p,
    ct.POINTER(ct.c_int),
    ct.POINTER(ct.c_char_p)
]

lib.py_extract_sparse.restype = ct.c_int
lib.py_extract_sparse.argtypes = [
    ct.c_void_p,
    ct.POINTER(ct.c_int),
    ct.POINTER(ct.c_char_p)
]

lib.py_free_mat.argtypes = [
    ct.c_void_p,
    ct.POINTER(ct.c_int),
    ct.POINTER(ct.c_char_p)
]

lib.py_initialize_compressed_sparse_matrix.restype = ct.c_void_p
lib.py_initialize_compressed_sparse_matrix.argtypes = [
    ct.c_int,
    ct.c_int,
    ct.c_uint64,
    ct.c_void_p,
    ct.c_void_p,
    ct.c_void_p,
    ct.c_void_p,
    ct.c_void_p,
    ct.c_uint8,
    ct.POINTER(ct.c_int),
    ct.POINTER(ct.c_char_p)
]

lib.py_initialize_dense_matrix.restype = ct.c_void_p
lib.py_initialize_dense_matrix.argtypes = [
    ct.c_int,
    ct.c_int,
    ct.c_void_p,
    ct.c_void_p,
    ct.c_uint8,
    ct.POINTER(ct.c_int),
    ct.POINTER(ct.c_char_p)
]

def extract_column(rawmat, c, output):
    return catch_errors(lib.py_extract_column)(rawmat, c, output)

def extract_ncol(mat):
    return catch_errors(lib.py_extract_ncol)(mat)

def extract_nrow(mat):
    return catch_errors(lib.py_extract_nrow)(mat)

def extract_row(rawmat, r, output):
    return catch_errors(lib.py_extract_row)(rawmat, r, output)

def extract_sparse(mat):
    return catch_errors(lib.py_extract_sparse)(mat)

def free_mat(mat):
    return catch_errors(lib.py_free_mat)(mat)

def initialize_compressed_sparse_matrix(nr, nc, nz, dtype, dptr, itype, iptr, indptr, byrow):
    return catch_errors(lib.py_initialize_compressed_sparse_matrix)(nr, nc, nz, dtype, dptr, itype, iptr, indptr, byrow)

def initialize_dense_matrix(nr, nc, type, ptr, byrow):
    return catch_errors(lib.py_initialize_dense_matrix)(nr, nc, type, ptr, byrow)